--
-- earein.e
-- ========
--
-- Source code beautifier / re-indent
-- ==================================
--
-- Invoked via Tools/Re-Indent source.
-- Processes the current source, inserting and removing 
--  leading whitespace (Positive/Eu source files only), 
--  concentrating mainly, but not exclusively, on the
--  if, while, and for statements.
-- Any changes it makes can be undone via Ctrl Z.
--
-- Notes. If a program relies on an external type (eg
--  global type bool(...) defined in another file), then
--  this program does not scan the included files (or the 
--  file(s) which include the current file!), and it must 
--  assume it is a type and display a warning.
--
-- Note1: One thing this will not cope with is:
--                          type fred(... end type
--                          procedure p(integer fred)
--                              fred=1
--  because we're not keeping a complete symbol table and
--  there is no record of the local definition of fred
--  overriding the type (or restoring it at end of scope).
--  Similarly for the builtin types, eg:
--                          integer integer integer=1
--
-- Particular sections of code can be wrapped in the
-- comment-style directives (beginning in column 1),
-- to prevent any reformatting between those points:
--
--#without reformat
--   ...
--#with reformat
--
--  There are a couple of examples of this being used
--  in edita.exw (version 0.2.9 or later).
--
-- There are four main parts to this program:
--
-- Part 1. Tokeniser and error routines.
-- Part 2. The main re-indent routine.
-- Part 3. The (highly simplified) parser.
-- Part 4. The main control routine.
--

--with trace

integer nest    -- main indent control; 0 for top-level, 4 in routines,
                -- +4 for if/for/while. NB tempered by Indent() parameter.
                -- (actually, it is isTabWidth rather than a literal 4)

sequence also   -- secondary indent control. Allows line-up of constants etc.
                -- (Can allow several permitted indent levels simultaneously)

integer reformat    -- controlled via --#with[out] reformat directive (in col 1).
                    -- normal reformatting when this is 1, none when it is 0.

integer lexer_tokno     -- token no in line as lexer(aka tokeniser) sees things
integer parser_tokno    -- token no in line as parser sees things (main)
                        -- (the lexer is usually one step ahead of whatever it
                        --  is we are currently parsing)

--
-- Part 1. Tokeniser and error routines
--

integer textlen
integer CurrLine, col, tokline, tokstart, tokend, toktype
sequence token

constant SYMBOL  = 1,
         DQUOTE  = 2,
         SQUOTE  = 3,
         ELLIPSE = 4,
         HEXDEC  = 5,
         DIGIT   = 6,
         FLOAT   = 7,
         USCORE  = 8,
         LETTER  = 9,
--       SLICE   = 10,
--       SUBSCR  = 11,
         ILLEGAL = 99

integer Ch

--integer errorcount
--      errorcount=0

sequence errtext

--DEV:: PARSE ERROR ON LINE NNN; JUMP TO LINE!
procedure Abort(sequence msg)
sequence t1
integer k
--  if errorcount=0 then
--DEV messageBox?/proemh?
    if tokline and tokline<=textlen then
--      printf(1,"%s\n%s^ %s.\n",{ExpandTabs(text[tokline]),
--                                repeat(' ',ExpLength(text[tokline][1..tokstart])-1),
--                                msg})
        t1=filetext[currfile][tokline]
        while 1 do
            k = find('\t',t1)
            if not k then exit end if
            t1[k]=' '
        end while
        CursorY = tokline-1
        errtext&=sprintf("line %d:\n%s\n%s^ %s.\n",{tokline,t1,repeat(' ',tokstart-1),msg})
--DEV use own dialogue, possibly fixed pitch, possibly repositioned.
--void = messageBox(sprintf("Error in line %d",{tokline}),
--            sprintf("%s\n%s^ %s.\n",{t1,repeat(' ',tokstart-1),msg}),0)
    else
        errtext&=sprintf("%s.\n",{msg})
--void = messageBox("Error",msg,0)
    end if

--      errorcount+=1
--  token=""
    Ch=-1
    toktype = ILLEGAL
    token="bogus"
end procedure

--with trace
procedure WarnType()--sequence fmt, object o)
--trace(1)
--  printf(1,fmt,o)
--  errtext&=sprintf(fmt,o)
    errtext&=sprintf(xl("Warning: %s assumed to be external type\n"),{token})
end procedure


procedure SkipSpacesAndComments()
    while 1 do
        while col=1 do
            if CurrLine>textlen then
                Ch =- 1
                return
            end if
            if col<=length(filetext[currfile][CurrLine]) then
                Ch = filetext[currfile][CurrLine][1]
                exit
            end if
            CurrLine += 1
            lexer_tokno = 1
        end while
        if Ch=' ' or Ch='\t' then
            col += 1
            if col>length(filetext[currfile][CurrLine]) then
                CurrLine += 1
                lexer_tokno = 1
                col = 1
            else
                Ch = filetext[currfile][CurrLine][col]
            end if
        elsif Ch='-'
          and col<length(filetext[currfile][CurrLine])
          and filetext[currfile][CurrLine][col+1]='-' then
            if col=1 and col+1<length(filetext[currfile][CurrLine])
            and filetext[currfile][CurrLine][col+2]='#' then
                if match("--#without reformat",filetext[currfile][CurrLine]) then
                    reformat = 0
                elsif match("--#with reformat",filetext[currfile][CurrLine]) then
                    reformat = 1
                end if
            end if
            CurrLine += 1
            lexer_tokno = 1
            col = 1
        else
            exit
        end if
    end while
end procedure

sequence charClass

--constant Operators={"<","<=",">",">=","=","!=",
--                  "+","+=","-","-=","*","*=",
--                  "/","/=","..","&","&="}
constant Operators={"<=",">=","!=","+=","-=","*=","/=","&="}
charClass=repeat(ILLEGAL,255)
charClass['!']=SYMBOL
charClass['\"']=DQUOTE
charClass['\'']=SQUOTE
charClass['#']=HEXDEC
charClass['&']=SYMBOL
charClass['('..'/']=SYMBOL  -- ()*+,-./
charClass['.']=ELLIPSE
charClass['0'..'9']=DIGIT
charClass[':'..'?']=SYMBOL  -- :;<=>?
charClass['_']=USCORE
charClass['A'..'Z']=LETTER
charClass['a'..'z']=LETTER
charClass['[']=SYMBOL
charClass[']']=SYMBOL
charClass['{']=SYMBOL
charClass['}']=SYMBOL


procedure nextCh()
    tokend = col-1
    if col<=length(filetext[currfile][CurrLine]) then
        Ch=filetext[currfile][CurrLine][col]
    else
        CurrLine+=1
        lexer_tokno = 1
        col=1
    end if
    SkipSpacesAndComments()
end procedure

procedure getFloat()
-- finish off a float. The first few DIGITS (if any) have been processed;
-- continue from '.' or 'eE' (we have no use for the value)
integer lenTC
    lenTC=length(filetext[currfile][CurrLine])
    if Ch='.' then
        -- make sure it is not an ellipse
        if col<lenTC
        and filetext[currfile][CurrLine][col+1] = '.' then
--          col-=1
        else
            toktype = FLOAT
            while 1 do
                col+=1
                if col>lenTC then exit end if
                Ch = filetext[currfile][CurrLine][col]
                if charClass[Ch] != DIGIT then exit end if
            end while           
        end if
    end if
    if find(Ch,"eE") 
    and col<lenTC then
        toktype = FLOAT
        Ch = filetext[currfile][CurrLine][col+1]
        if Ch = '-' or Ch = '+' then
            col+=1
            Ch = filetext[currfile][CurrLine][col+1]
        end if
        while 1 do
            col+=1
            if col>lenTC then exit end if
            Ch = filetext[currfile][CurrLine][col]
            if charClass[Ch] != DIGIT then exit end if
        end while
    end if
    nextCh()
end procedure

--function escMap(integer escChar)
--  if escChar='t' then return '\t' end if
--  if escChar='r' then return '\r' end if
--  if escChar='n' then return '\n' end if
--  if escChar='\"' then return '\"' end if
--  if escChar='\'' then return '\'' end if
--  if escChar='\\' then return '\\' end if
--  -- oops!
--end function

integer mapEndToMinusOne
        mapEndToMinusOne=0

--with trace
procedure getToken()
integer nxtCh, lenTC
    if Ch = -1 then return end if
    tokline = CurrLine
    tokstart = col
    parser_tokno = lexer_tokno
    lexer_tokno += 1
    token = ""
--if Ch='?' then trace(1) end if
    toktype = charClass[Ch]
    lenTC = length(filetext[currfile][CurrLine])
    if col < lenTC then
        nxtCh = filetext[currfile][CurrLine][col+1]
    else
        nxtCh = -1
    end if
    if toktype = LETTER then
        while 1 do
            col += 1
            if col > lenTC then exit end if
            nxtCh = charClass[filetext[currfile][CurrLine][col]]
            if nxtCh < DIGIT then exit end if
            if nxtCh > LETTER then exit end if
        end while
        token = filetext[currfile][CurrLine][tokstart..col-1]
        if mapEndToMinusOne and equal(token,"end") then
            token = "-1"
            toktype = DIGIT
        end if
        nextCh()
    elsif toktype = SYMBOL then
        if find({Ch,nxtCh},Operators) then
            col += 1
        elsif Ch = '?' then
            toktype = LETTER
        end if
        col += 1
        token = filetext[currfile][CurrLine][tokstart..col-1]
        nextCh()
    elsif toktype = DQUOTE then
        while 1 do
            col += 1
            if col>lenTC then
                Abort(xl("missing \""))
                exit
            end if
            nxtCh = filetext[currfile][CurrLine][col]
            if nxtCh = '\\' then
                col += 1
            elsif nxtCh = '\"' then
                col += 1
                exit
            end if
        end while
        token = filetext[currfile][CurrLine][tokstart..col-1]
        nextCh()
    elsif toktype = SQUOTE then
        if nxtCh = '\\' then
--          tokint = escMap(filetext[currfile][CurrLine][col+2])
            col += 4
        else
--          tokint = filetext[currfile][CurrLine][col+1]
            col += 3
        end if
--      if col>lenTC then ?9/0 end if
--      token = filetext[currfile][CurrLine][tokstart..col-1]
--      col+=1
        nextCh()
    elsif toktype = HEXDEC then
        while 1 do
            col += 1
            if col>lenTC then exit end if
            nxtCh = filetext[currfile][CurrLine][col]
            if nxtCh < '0' then exit end if
            if nxtCh > '9' then 
                if nxtCh < 'A' then exit end if
                if nxtCh > 'F' then
                    if nxtCh < 'a' then exit end if
                    if nxtCh > 'f' then exit end if
                end if
            end if
        end while
        nextCh()
    elsif toktype = ELLIPSE then
        if nxtCh = '.' then
            col += 2
            nextCh()
        else
            getFloat()
        end if
    elsif toktype <= DIGIT then
        while 1 do
            col += 1
            if col > lenTC then exit end if
            Ch = filetext[currfile][CurrLine][col]
            if charClass[Ch] != DIGIT then exit end if
        end while
        if find(Ch,".eE") then
            getFloat()
        else
            nextCh()
        end if
    elsif toktype = ILLEGAL 
      and mapEndToMinusOne 
      and Ch = '$' then
        token = "-1"
        toktype = DIGIT
        col += 1
        nextCh()
    else
        Abort(xl("unrecognised"))
    end if
end procedure

procedure Check(sequence text)
    if not equal(token,text) then
        Abort(text&xl(" expected"))
    end if
end procedure

procedure Match(sequence text)
    if not equal(token,text) then
        Abort(text&xl(" expected"))
    end if
    getToken()
end procedure

constant testharness = (routine_id("paintall")=-1)
--if testharness then:
function expandtabs(sequence text)
integer tab
    while 1 do
        tab=find('\t',text)
        if not tab then exit end if
        text=text[1..tab-1]&
             repeat(' ',isTabWidth-remainder(tab-1,isTabWidth))&
             text[tab+1..length(text)]
    end while
    return text
end function


--
-- Part 2. The main re-indent routines
--
integer adjust_count, 
        nooftabs, 
        noofspaces

integer lastIline, lastIaction, lastIcount
constant indent=1, un_indent=-1
--if testharness then:
sequence un

procedure completeindents()
    if lastIline!=-1 then
        CursorY = lastIline-1
        selX = 0
        CursorX = 0
        if testharness then
            un = ""
            if lastIaction=-1 then
                un = "un"
            end if
            printf(1,"%sindent[%d] lines %d..%d\n",
                   {un,lastIcount,selY+1,lastIline})
--if getc(0) then end if
        end if
--printf(1,"%sindent[%d] lines %d..%d\n",{{"","un"}[lastIaction],lastIcount,selY+1,lastIline})
        for i=1 to lastIcount do
            if lastIaction = indent then
                addAction(INDENT,{VK_TAB,indentWith(VK_TAB)})
            else
                addAction(UNINDENT,{VK_TAB,unindent(VK_TAB)})
            end if
        end for
        adjust_count += 1
        lastIaction = 0
        lastIline = -1
        selON=0
    end if
end procedure

function inPend()
    if tokline!=lastIline then return 0 end if
    return lastIaction*lastIcount*isTabWidth
end function

procedure dragCommentsAlong(integer direction)
sequence wsmatch -- whitespace + comment ("--"); required match string.
    if (direction = 1 and tokline>lastIline+1)
    or (direction =-1 and selY) then
        wsmatch = filetext[currfile][lastIline]
        for k=1 to length(wsmatch) do
            if not find(wsmatch[k]," \t") then
                wsmatch = wsmatch[1..k-1]
                exit
            end if
        end for
        wsmatch &= "--"
        if direction = 1 then
            while tokline>lastIline+1
              and (match(wsmatch,filetext[currfile][lastIline+1])=1
                   or length(filetext[currfile][lastIline+1])=0) do
                lastIline+=1
            end while
        else
            while selY
              and (match(wsmatch,filetext[currfile][selY])=1
                   or length(filetext[currfile][selY])=0) do
                selY -= 1
            end while
        end if
    end if
end procedure

procedure indentline()
    if lastIline!=-1 then
        dragCommentsAlong(1)
        if tokline!=lastIline+1
        or lastIaction!=indent
        or lastIcount!=nooftabs then
            completeindents()
        end if
    end if
    lastIaction = indent
    lastIcount = nooftabs
    lastIline = tokline
    if selON=0 then
        selY = tokline-1
        selON = 1
        dragCommentsAlong(-1)
    end if
end procedure

function unindentline(integer w, sequence line)
    for i=1 to w do
        if line[i]!='\t' then return 0 end if
    end for
    if lastIline!=-1 then
        dragCommentsAlong(1)
        if tokline != lastIline+1
        or lastIaction != un_indent
        or lastIcount != w then
            completeindents()
        end if
    end if
    lastIaction = un_indent
    lastIcount = w
    lastIline = tokline
    if selON=0 then
        selY = tokline-1
        selON = 1
        dragCommentsAlong(-1)
    end if
    return 1
end function

integer newRoutineBlock, wasNewRoutineBlock, undoExtraNest
        newRoutineBlock=0 undoExtraNest=0

procedure Indent(integer i)
-- check that line tokline begins with correct whitespace
sequence line, tabwork
--if testharness then:
, prevline
integer w, ch
    wasNewRoutineBlock = newRoutineBlock
    newRoutineBlock = 0
    if not reformat then return end if --#without reformat in force.
    line = filetext[currfile][tokline]
    i += nest
    w = 0
    for j = 1 to length(line) do
        ch = line[j]
        if ch = '\t' then
            w += isTabWidth - remainder(w, isTabWidth)
        elsif ch = ' ' then
            w += 1
        else
            exit
        end if
    end for
    w = i-w
    if length(also) then
--      if i=w then
--          w = 0
--      else
        for j=1 to length(also) do
            if i=also[j] or i=also[j]+w then
                w = 0
                exit
            end if
        end for
--      end if
    end if
    if w then
        if testharness then
            prevline = line
        end if
        for j = 1 to length(line) do
            if not find(line[j]," \t") then
--              trace(1)
                nooftabs = floor(i/isTabWidth)
                noofspaces = i-nooftabs*isTabWidth
                w = j-1
                while w and noofspaces and line[w]=' ' do
                    noofspaces -= 1
                    w -= 1
                end while
                if noofspaces = 0 then
                    if j<=length(line)-4 then
                        --
                        -- This permits expressions such as:
                        --      elsif xxx
                        --        and xxx
                        --         or xxx
                        -- rather than reformatting them to:
                        --      elsif xxx
                        --      and xxx
                        --      or xxx
                        -- This code itself is an example for Params()'s () handling.
                        --
                        if (    line[j]  ='o'
                            and line[j+1]='r'
                            and find(line[j+2]," \t"))
                        or (    line[j]  ='a'
                            and line[j+1]='n'
                            and line[j+2]='d'
                            and find(line[j+3]," \t")) then
--trace(1)
                            if j-nooftabs-1<=3 then
                                for k=1 to j-1 do
                                    if k<=nooftabs then
                                        if line[k]!='\t' then exit end if
                                    else
                                        if line[k]!=' ' then exit end if
                                        if k = j-1 then return end if   
                                    end if
                                end for
                            end if
                        end if
                    end if
--trace(1)
                    while w and nooftabs and line[w]='\t' do
                        nooftabs -= 1
                        w -= 1
                    end while
                    if w = 0 then   -- pure indent
                        indentline()
                        return
                    else            -- pure un_indent?
                        if wasNewRoutineBlock then
                            -- allow either:
                            --      procedure x()
                            --          a=0
                            --          b=0
                            --      end procedure
                            -- or:
                            --      procedure x()
                            --              a=0
                            --              b=0
                            --      end procedure
                            --
                            if w=1 then
                                undoExtraNest = 1
                                nest += isTabWidth
                                return
                            end if
                        end if
                        if unindentline(w,line) then return end if
                    end if
                end if
                completeindents()
                tabwork = {repeat('\t',nooftabs)&
                           repeat(' ',noofspaces)}
--              tabwork = {repeat(' ',i)}
                if not testharness then
                    selON=1
                    CursorY=tokline-1
                    selY=CursorY
                    selX=0
                    CursorX=ExpLength(line[1..w])
                    line = {}           -- reduce ref count
                    if deleteSelection() then end if
                    if nooftabs+noofspaces then
                        addAction(INSERTBLOCK,tabwork)
                        InsertBlock(tabwork)
                    end if
                else
                    line = tabwork[1] & line[w+1..length(line)]
                    filetext[currfile][tokline] = line
                end if
                if tokline = CurrLine then
                    w -= nooftabs+noofspaces
                    col -= w
                    tokstart -= w   -- for DoIf
                end if
                exit
            end if
        end for
        if testharness then
            printf(1,"%s[line %d]=>\n%s\n",{expandtabs(prevline),tokline,expandtabs(line)})
            if getc(0) then end if
        end if
        adjust_count += 1
    end if
end procedure


--
-- Part 3. The (highly simplified) parser.
--

integer rExpression

procedure DoSequence()
integer wasNest -- see eg Expression()
    wasNest = nest
--  nest = ExpLength(filetext[currfile][tokline][1..tokstart-1])+inPend()
    nest = ExpLength(filetext[currfile][tokline][1..tokstart])+inPend()
    Match("{")
    if parser_tokno=1 then Indent(0) end if
    while not equal(token,"}") do
        call_proc(rExpression,{})
        if not equal(token,",") then exit end if
        Match(",")
        if parser_tokno=1 then Indent(0) end if
    end while
    if parser_tokno=1 then nest-=1 Indent(0) end if
    nest = wasNest
    Match("}")
end procedure

procedure Params()
integer wasNest -- see eg Indent()
    getToken()
    Match("(")
    if not equal(token,")") then
        wasNest = nest
        nest = ExpLength(filetext[currfile][tokline][1..tokstart-1])+inPend()
        while 1 do
            call_proc(rExpression,{})
            if not equal(token,",") then exit end if
            getToken()
            if parser_tokno=1 then Indent(0) end if
        end while
        nest = wasNest
    end if
    Match(")")
end procedure

procedure DoSubScripts()
integer wasMapEndToMinusOne
    wasMapEndToMinusOne = mapEndToMinusOne
    mapEndToMinusOne = 1
    while Ch = '[' do
        getToken()
        Match("[")
        call_proc(rExpression,{})
        if toktype = ELLIPSE then
            getToken()
            call_proc(rExpression,{})
            Check("]")
            exit
        end if
        Check("]")
    end while
    mapEndToMinusOne = wasMapEndToMinusOne
    getToken()
end procedure

procedure Factor()
integer wasNest
    if toktype = SYMBOL and find(token,{"+","-"}) then
        Match(token)
    elsif toktype = LETTER and equal(token,"not") then
        Match(token)
    end if
    if toktype = LETTER then
        if Ch = ':' then getToken() getToken() end if -- skip namespaces
        if Ch = '(' then    -- a function, we presume
            Params()
        else            -- a variable, we presume
            DoSubScripts()
        end if
    elsif find(toktype,{DIGIT,FLOAT,HEXDEC,DQUOTE,SQUOTE}) then
        getToken()
    elsif toktype = SYMBOL and equal(token,"{") then
        DoSequence()
    elsif toktype = SYMBOL and equal(token,"(") then
        wasNest = nest
        nest = ExpLength(filetext[currfile][tokline][1..tokstart])+inPend()
        getToken()
        call_proc(rExpression,{})
        Match(")")
        nest = wasNest
    else
        Abort(xl("unrecognised"))
    end if
end procedure

--procedure NotFactor()
--  if toktype = SYMBOL and find(token,{"+","-"}) then
--      Match(token)
--  elsif toktype = LETTER and equal(token,"not") then
--      Match(token)
--  end if
--  Factor()
--end procedure
--
--procedure Term()
--  NotFactor()
--  while find(token,{"*","/"}) do
--      Match(token)
--      NotFactor()
--  end while
--end procedure
--
--procedure Sum()
--  Term()
--  while find(token,{"+","-"}) do
--      Match(token)
--      Term()
--  end while
--end procedure
--
--procedure Concat()
--  Sum()
--  while equal(token,"&") do
--      Match("&")
--      Sum()
--  end while
--end procedure
--
--procedure Relation()
--  Concat()
--  while find(token,{"<","<=","=","!=",">",">="}) do
--      Match(token)
--      Concat()
--  end while
--end procedure

procedure Expression()
--  Relation()
    Factor()
--  while find(token,{"and","or","xor"}) do
    while find(token,{"and","or","xor",
                      "<","<=","=","==","!=",">",">=",
                      "&",
                      "+","-",
                      "*","/"}) do
        if parser_tokno=1 then Indent(0) end if
        Match(token)
        if parser_tokno=1 then Indent(0) end if
--      Relation()
        Factor()
    end while
end procedure
rExpression=routine_id("Expression")

integer rBlock

procedure DoIf(integer returnexpr)
--
-- Recognize and Translate an IF Construct
-- returnexpr is a flag, 1 for function/type return <expr>,
--                       0 for procedure returns,
--                      -1 for toplevel statements
--
integer wasNest
object wasalso
    wasNest = nest
    wasalso = also
    if tokstart>1 or nest or inPend() then
        nest += isTabWidth
    end if
    Match("if")
    also = {ExpLength(filetext[currfile][tokline][1..tokstart-1])+inPend()}
    Expression()
    also = {}
    Match("then")
    call_proc(rBlock,{returnexpr})
    while equal(token,"elsif") do
        if parser_tokno=1 then Indent(0) end if
        Match("elsif")
        Expression()
        Match("then")
        call_proc(rBlock,{returnexpr})
    end while
    if equal(token,"else") then
        if parser_tokno=1 then Indent(0) end if
        Match("else")
        call_proc(rBlock,{returnexpr})
    end if
    if parser_tokno=1 then Indent(0) end if
    nest = wasNest
    Match("end")
    Match("if")
    also = wasalso
end procedure

procedure DoWhile(integer returnexpr)
--
-- Parse and Translate a WHILE Statement
-- returnexpr is a flag, 1 for function/type return <expr>,
--                       0 for procedure returns,
--                      -1 for toplevel statements
--
integer wasNest
    wasNest = nest
    if tokstart>1 or nest or inPend() then
        nest += isTabWidth
    end if
    Match("while")
    Expression()
    Match("do")
    call_proc(rBlock,{returnexpr})
    if parser_tokno=1 then Indent(0) end if
    nest = wasNest
    Match("end")
    Match("while")
end procedure

procedure DoFor(integer returnexpr)
----
---- Parse and Translate a For Statement
---- returnexpr is a flag, 1 for function/type (permitting return <expr>),
----                         0 for procedure (permitting return),
----                        -1 for toplevel statements (prohibiting return),
----                           and forcing control var to be static, not stack.
----
integer wasNest
    wasNest = nest
    if tokstart>1 or nest or inPend() then
        nest += isTabWidth
    end if
    Match("for")
    getToken()
    Match("=")
    Expression()
    Match("to")
    Expression()
    if equal(token,"by") then
        getToken()
        Expression()
    end if
    Match("do")
    call_proc(rBlock,{returnexpr})
    if parser_tokno=1 then Indent(0) end if
    nest = wasNest
    Match("end")
    Match("for")
end procedure

procedure DoExit()
    Match("exit")
end procedure

procedure DoReturn(integer returnexpr)
---- returnexpr is a flag, 1 for function/type return <expr>,
----                         0 for procedure returns,
----                        -1 for toplevel statements
integer wasNest
    if returnexpr=-1 then Abort(xl("return must be inside a procedure or function")) end if
    Match("return")
    if returnexpr then
        wasNest = nest
        nest = ExpLength(filetext[currfile][tokline][1..tokstart-1])+inPend()
        Expression()
        nest = wasNest
    end if
end procedure

procedure DoQu()
----
---- The '?' shorthand.
----
    getToken()
    Expression()
end procedure

procedure Assignment()
integer wasNest
    DoSubScripts()
    if not find(token,{"=","+=","-=","*=","/=","&="}) then
        Abort(xl("assignment operator expected"))
    else
        getToken()
        wasNest = nest
        nest = ExpLength(filetext[currfile][tokline][1..tokstart-1])+inPend()
        Expression()
        nest = wasNest
    end if
end procedure

sequence vartypes
procedure DoRoutineDef(integer rType)
sequence sType, tType

    sType = token   -- procedure/function/type
    getToken()
    tType = token   -- type name
    getToken()
    Match("(")
    if not equal(token,")") then
        while 1 do
            if not find(token,vartypes) then    -- see Note1
                if charClass[Ch] != LETTER then
                    Abort(xl("a type is expected here"))
                    exit
                end if
--              Warn("Warning: %s assumed to be external type\n",{token})
                WarnType()
                vartypes = append(vartypes,token)
            end if
            getToken()
            getToken()
            if not equal(token,",") then exit end if
            getToken()
--DEV positive types..
--          while Ch=',' do
--              getToken()
--              getToken()
--          end while
        end while
    end if
    Match(")")
    also = {0,4}
    while toktype = LETTER do
        if not find(token,vartypes) then    -- see Note1
            if find(token,{"if","for","while","return","end"}) then exit end if
            --
            if charClass[Ch] != LETTER then exit end if
            -- ^explanation^: word word (not word =, or word[... etc) /must/ be
            --                a variable definition (eg boolean t), not code.
            --
--          Warn("Warning: %s assumed to be external type\n",{token})
            WarnType()
            vartypes = append(vartypes,token)
        end if
        if parser_tokno=1 then Indent(0) end if
        while 1 do
            getToken()
            if Ch!=',' then exit end if
            getToken()
        end while
        getToken()
    end while
    also = {}
    newRoutineBlock=1
    call_proc(rBlock,{rType>1})
    if undoExtraNest then
        undoExtraNest = 0
        nest -= isTabWidth
    end if
    newRoutineBlock=0   -- just in case
    also = {0,4}
    if parser_tokno=1 then Indent(0) end if
    Match("end")
    Match(sType)
    if rType = 3 then
        vartypes = append(vartypes,tType)
    end if
end procedure

procedure Statement(integer returnexpr)
integer switch
    if parser_tokno=1 then Indent(isTabWidth) end if
    if toktype = LETTER then
        switch = find(token,{"if","for","while","exit","return","?"})
        if switch = 1 then DoIf(returnexpr)
        elsif switch = 2 then DoFor(returnexpr)
        elsif switch = 3 then DoWhile(returnexpr)
        elsif switch = 4 then DoExit()
        elsif switch = 5 then DoReturn(returnexpr)
        elsif switch = 6 then DoQu()
        elsif Ch='(' then Params()
        elsif Ch=':' then getToken() getToken() -- skip namespaces
        else Assignment()
        end if
    else
        Abort(xl("unrecognised"))
    end if
end procedure

procedure Block(integer returnexpr)
    while toktype = LETTER do
        if find(token,{"elsif","else","end"}) then exit end if
        Statement(returnexpr)
    end while
end procedure
rBlock=routine_id("Block")

procedure TopDecls()
    if parser_tokno=1 then Indent(0) end if
    while 1 do
        getToken()
        if Ch!=',' then exit end if
        getToken()
    end while
    getToken()
end procedure

procedure DoConstant()
    getToken()  -- discard "constant".
    while 1 do
        if parser_tokno=1 then Indent(0) end if
        getToken()          -- constant name
        Match("=")
        if Ch=':' then getToken() getToken() end if -- skip namespaces
        Expression()
        if not equal(token,",") then exit end if
        getToken()
    end while
end procedure

--
-- Part 4. The main control routine.
--

global procedure ReIndent()
integer t
sequence aline

    CurrLine = 1
    lexer_tokno = 1
    adjust_count = 0
    reformat = 1
    lastIline = -1
    nest = 0
    errtext = ""
    also = {0}
    vartypes={"atom","integer","sequence","object"}
    textlen = length(filetext[currfile])
    if textlen
    and match("#!",filetext[currfile][1])=1 then    -- skip shebang
        CurrLine = 2
    end if
    col = 1
    SkipSpacesAndComments()
    getToken()
    while Ch!=-1 do
        if toktype = LETTER then
            t = find(token,{"procedure","function","type"})
            if t then
                also = {}
                if parser_tokno=1 then Indent(0) end if
                DoRoutineDef(t)
                also = {0}
            elsif equal(token,"constant") then
                also = {ExpLength(filetext[currfile][CurrLine][1..col-1])}
                if parser_tokno=1 then Indent(0) end if
                DoConstant()
                if not find(0,also) then also = prepend(also,0) end if
            elsif equal(token,"global") then
                also = {}
                if parser_tokno=1 then Indent(0) end if
                getToken()  -- otherwise ignore it
            elsif find(token,{"include","with","without"}) then
                also = {}
                if parser_tokno=1 then Indent(0) end if
                CurrLine+=1
                col=1
                SkipSpacesAndComments()
                getToken()
                also = {0}
--          elsif equal(token,"?") then
--              DoQu()
            else
                if Ch=':' then      -- eg a:fred
                    getToken()      -- skip "a"
                    getToken()      -- skip ":", leave on "fred"
                end if -- skip namespaces
                t = find(token,vartypes)
                if not t and not equal(token,"?") then
                    if charClass[Ch] = LETTER
                    and not find(token,{"if","for","while"}) then
                        WarnType()
                        vartypes = append(vartypes,token)
                        t = length(vartypes)
                    end if
                end if
                if t then
                    if parser_tokno=1 then Indent(0) end if
                    also = {ExpLength(filetext[currfile][CurrLine][1..col-1])}
                    TopDecls()
                    if not find(0,also) then also = prepend(also,0) end if
                else
                    if equal(token,"abort") and Ch='(' then
                        -- stop processing, and warn if remainder of file
                        -- is other than just blank lines and comments
                        for j=CurrLine+1 to length(filetext[currfile]) do
                            aline = filetext[currfile][j]
                            while length(aline) and find(aline[1]," \t\r\n") do
                                aline = aline[2..length(aline)]
                            end while
                            if length(aline)!=0 and match("--",aline)!=1 then
-- this is not an additional prompt, just extra info on the normal result window:
errtext&=sprintf(xl("Warning: top-level abort on line %d; no further lines processed\n"),{CurrLine})
                                exit
                            end if
                        end for
                        exit
                    end if
                    Statement(-1)
                end if
            end if
        else
            Abort(xl("invalid"))
            exit
        end if
-- Not worth doing this, really:
--      doEvents(0)
    end while
    completeindents()
    forceCursorOnscreen()
    ensureVisible(CursorY,0)
    errtext &= sprintf(xl("Re-Indent complete: %d change%s made"),
                       {adjust_count,repeat('s',adjust_count!=1)})
    void = proemh(xl("ReIndent"), errtext, 0)
end procedure


